/*
This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/

#include <stdio.h> //printf debugging
#include "btWheelShapes.h"
#include "BulletCollision/CollisionShapes/btConeShape.h"
#include "BulletCollision/CollisionShapes/btConvexHullShape.h"

static btVector3 GetAabbFullExtents(const btCollisionShape* shape)	{
    if (shape)	{
        btVector3 aabbMin,aabbMax;
        shape->getAabb(btTransform::getIdentity(),aabbMin,aabbMax);
        return (aabbMax - aabbMin);
    }
    return btVector3(0,0,0);
}
static btVector3 GetAabbHalfExtents(const btCollisionShape* shape)	{
    return GetAabbFullExtents(shape)*btScalar(0.5);
}


// From <btCylinderShape.cpp>: added optional cylinderHollowRadius (but not used AFAIK)
static void CalculateCylinderLocalInertia(btScalar cylinderRadius,btScalar cylinderHeight,btScalar mass,btVector3& localInertia,btScalar cylinderHollowRadius = btScalar(0),int cylinderHeightAxis=0)
{

    /*
    cylinder is defined as following:
    *
    * - principle axis aligned along y by default, radius in x, z-value not used
    * - for btCylinderShapeX: principle axis aligned along x, radius in y direction, z-value not used
    * - for btCylinderShapeZ: principle axis aligned along z, radius in x direction, y-value not used
    *
    */

    btScalar radius2;	// square of cylinder radius = (cylinderRadius*cylinderRadius + cylinderHollowRadius+cylinderHollowRadius)
    btScalar height2;	// square of cylinder height
    btScalar div12 = mass / 12.f;
    btScalar div4 = mass / 4.f;
    btScalar div2 = mass / 2.f;
    int idxRadius, idxHeight;

    switch (cylinderHeightAxis)	// get indices of radius and height of cylinder
    {
        case 0:		// cylinder is aligned along x
            idxRadius = 1;
            idxHeight = 0;
            break;
        case 2:		// cylinder is aligned along z
            idxRadius = 0;
            idxHeight = 2;
            break;
        default:	// cylinder is aligned along y
            idxRadius = 0;
            idxHeight = 1;
    }

    // calculate squares
    radius2 = cylinderRadius * cylinderRadius + (cylinderHollowRadius>0 ? cylinderHollowRadius * cylinderHollowRadius : cylinderHollowRadius<0 ? -cylinderHollowRadius*cylinderHollowRadius : 0);
    height2 = cylinderHeight * cylinderHeight;

    // calculate tensor terms
    btScalar t1 = div12 * height2 + div4 * radius2;
    btScalar t2 = div2 * radius2;

    switch (cylinderHeightAxis)	// set diagonal elements of inertia tensor
    {
        case 0:		// cylinder is aligned along x
            localInertia.setValue(t2,t1,t1);
            break;
        case 2:		// cylinder is aligned along z
            localInertia.setValue(t1,t1,t2);
            break;
        default:	// cylinder is aligned along y
            localInertia.setValue(t1,t2,t1);
    }

}


void btWheelCylinderShape::setExtendedCollisionMargin() {
    const btVector3 aabbHalfExtents = GetAabbHalfExtents(this);
    const btScalar radius = aabbHalfExtents.y();
    const btScalar height = aabbHalfExtents.x()*2.0f;
    this->setMargin(radius*0.5f);   // ???
}

btMultiSphereShape* btWheelSimpleShape::Create(const btScalar radius,const btScalar height)	{
	
	const btScalar scaling = height/(radius*2.0f);	
	const btVector3 positions[] = { btVector3(0.f,0.f,0.f)};
	const btScalar radii[] = {	radius };
	const btVector3	localScaling = btVector3(scaling,1.0f,1.0f);						
	const int numSpheres = sizeof(positions)/sizeof(positions[0]);							
	if (numSpheres != sizeof(radii)/sizeof(radii[0])) {
		printf("Error: numPositions!=numRadii in btMultiSphereShape [btWheelSimpleShape::Create(...)]\n");
		return NULL;
	}							
	btMultiSphereShape* shape = new btMultiSphereShape(&positions[0],&radii[0],numSpheres);
	shape->setLocalScaling(localScaling);
		
	return shape;
}
void btWheelSimpleShape::calculateLocalInertia(btScalar mass,btVector3& inertia) const    {
    btScalar radius = -1;
    btScalar height = -1;

    const btVector3 aabbHalfExtents = GetAabbHalfExtents(this);
    radius = aabbHalfExtents.y();
    height = aabbHalfExtents.x()*2.0f;

    if (radius<=0 || height<=0) base::calculateLocalInertia(mass,inertia);
    else CalculateCylinderLocalInertia(radius,height,mass,inertia);
}

btMultiSphereShape* btWheelShape::Create(const btScalar radius,const btScalar height,const int numSegments,const btScalar _smallSpheresHeightFactor)	{
	//const btScalar height = _height*0.35f;	// Empiric fix
	
	const int numOfSmallSpheresPerCircle = numSegments;	// 12 // Twickable (less = faster, higher = more precise)
	const btScalar smallSpheresHeightFactor = _smallSpheresHeightFactor;//0.95f;	// (0,1) I'd stick to 0.95f (lower values should increase the tyre curvature at the center)
	
	const btScalar scaling = height/(radius*2.0);//height/(radius*2.0);	// correct value: height/(radius*2.0); we don't use it to have a better round tyre surface towards the ground (exceeding a bit the shape height (i.e.width))	
	const btScalar newHeight = radius*2.0;	// When we'll scale it, it will became = height
	
	const int numPositions = 2*numOfSmallSpheresPerCircle+1;	// This is the number of spheres in the btMultiSphereShape
	btAlignedObjectArray<btVector3> positions;positions.resize(numPositions);
	btAlignedObjectArray<btScalar> radii;radii.resize(numPositions);
	
	int i = 0;
	positions[i] = btVector3(0.f,0.f,0.f);
	radii[i++] = radius;
	
	// Small spheres
	const btScalar delta = smallSpheresHeightFactor * radius;	// + smallRadius
	const btScalar smallRadius = newHeight*0.25 < delta*0.5 ? newHeight*0.25 : delta*0.5;
	const btScalar R= delta-smallRadius >= smallRadius ? delta-smallRadius : smallRadius;
	const btScalar h1 = newHeight*0.25*0.75;	// The "correct" value should be newHeight*0.25, but the 0.75 factor increases the tyre curvature at the borders a bit, making it more realistic (TODO: further test this).
	const btScalar h2 = -h1;
	
	for (int t=0;t<numOfSmallSpheresPerCircle;t++)	{
		const btScalar s = R*btSin(2.f*SIMD_PI*(btScalar)t/numOfSmallSpheresPerCircle);
		const btScalar c = R*btCos(2.f*SIMD_PI*(btScalar)t/numOfSmallSpheresPerCircle);
		positions[i] = btVector3(h1,s,c);
		radii[i++] = smallRadius;
		positions[i] = btVector3(h2,c,s);
		radii[i++] = smallRadius;
	}
		
	const btVector3	localScaling = btVector3(scaling,1.0,1.0);						
	const int numSpheres = numPositions;//sizeof(positions)/sizeof(positions[0]);							
	if (numSpheres != radii.size()) {
		printf("Error: numPositions!=numRadii in btMultiSphereShape [btWheelShape::Create(...)]\n");
		return NULL;
	}							
	btMultiSphereShape* shape = new btMultiSphereShape(&positions[0],&radii[0],numSpheres);
	shape->setLocalScaling(localScaling);
	
	return shape;
}
void btWheelShape::calculateLocalInertia(btScalar mass,btVector3& inertia) const    {
    btScalar radius = -1;
    btScalar height = -1;

    const btVector3 aabbHalfExtents = GetAabbHalfExtents(this);
    radius = aabbHalfExtents.y();
    height = aabbHalfExtents.x()*2.0f;

    if (radius<=0 || height<=0) base::calculateLocalInertia(mass,inertia);
    else CalculateCylinderLocalInertia(radius,height,mass,inertia);
}

btCompoundShape* btWheelTrainShape::Create(btScalar outerRadius,btScalar totalWidth,btScalar _innerRadius,btScalar _innerWidth)	{
	const btScalar defaultParametersFactor = 0.7f/1.4f;//0.6f/1.4f;	// inner / outer dimensions when using default params
	btScalar innerRadius(_innerRadius);if (innerRadius<0 || innerRadius>=outerRadius) innerRadius = outerRadius*defaultParametersFactor;
	btScalar innerWidth(_innerWidth);if (innerWidth<0 || innerWidth>=totalWidth) innerWidth = totalWidth*defaultParametersFactor;
	//printf("outerRadius = %10.2f\ntotalWidth = %10.2f\nInner Width = %10.2f\nInner Radius = %10.2f\n",outerRadius,totalWidth,innerWidth,innerRadius);	
	const btScalar outerWidth=(totalWidth - innerWidth)*0.5f;
	
	// If the 2 apicies of the cones don't come out of the side they're facing we can use 2 cones and a cylinder:
		const btScalar deltaRadius = outerRadius-innerRadius; // DR : OW = OR : TW | TW = OW * OR / DR 
		const btScalar coneHeight =  outerWidth * outerRadius / deltaRadius;
		if (totalWidth<coneHeight) totalWidth = coneHeight;	// Yes, we override the totalWidth argument passed by the user to fulfil the condition below. 
															// P.S. It's surely possible to delete this requirement, but this code is old, and I'm very lazy... do it better yourself!
		//if (totalWidth>=coneHeight) {
			btCylinderShape* centralShape = new btCylinderShapeX(btVector3(totalWidth*0.5,innerRadius,innerRadius));//(btVector3(innerWidth*0.5,innerRadius,innerRadius));	// We replace innerWidth with totalWidth to make it bigger and overlapping
			btConeShape* coneShape = new btConeShape(outerRadius,coneHeight);			
			// Compound Shape
    		btCompoundShape* wheelShape = new btCompoundShape();
            wheelShape->addChildShape(btTransform::getIdentity(),centralShape);
			const btQuaternion Q1(btVector3(0.f,0.f,1.f),-SIMD_HALF_PI);
			wheelShape->addChildShape(btTransform(Q1,btVector3(-totalWidth*0.5+coneHeight*0.5,0,0)),coneShape);
			const btQuaternion Q2(btVector3(0.f,0.f,1.f),SIMD_HALF_PI);
			wheelShape->addChildShape(btTransform(Q2,btVector3(+totalWidth*0.5-coneHeight*0.5,0,0)),coneShape);			
		
			return wheelShape;
		//}
}
void btWheelTrainShape::calculateLocalInertia(btScalar mass,btVector3& inertia) const    {
    btScalar radius = -1;
    btScalar height = -1;

    const btVector3 aabbHalfExtents = GetAabbHalfExtents(this);
    height = aabbHalfExtents.x()*2.0f;
    radius = aabbHalfExtents.y();

    //--------------------------------------------------------------------------------------
    btScalar hollowRadius = 0;
    // This part is extremely experimental: feel free to comment it out...
    if (this->getNumChildShapes()>0 && this->getChildShape(0) && this->getChildShape(0)->getShapeType()==CYLINDER_SHAPE_PROXYTYPE) hollowRadius = - (radius - GetAabbHalfExtents(this->getChildShape(0)).y());  //We set it negative on purpose, cause the train wheel has an "external hollow part"
    //radius*=btScalar(0.75);   // This could be a simple replecement of the line above
    //---------------------------------------------------------------------------------------
    //REASON:
    /*
    The train wheel shape is not hollow at all: it is = "big cylinder" - "approximated hollow cylinder":
    so we should reduce the "big radius" somehow to calculate a better local inertia
    */
    //---------------------------------------------------------------------------------------

    if (radius<=0 || height<=0) base::calculateLocalInertia(mass,inertia);
    else CalculateCylinderLocalInertia(radius,height,mass,inertia,hollowRadius);
}



#define USE_BTTRIANGLEMESHEX
#ifdef  USE_BTTRIANGLEMESHEX
#include <GlutDemoApplicationEx.h>
#else //USE_BTTRIANGLEMESHEX
typedef void btTriangleMeshEx;
typedef void GlutDemoApplicationEx;
#endif //USE_BTTRIANGLEMESHEX

SIMD_FORCE_INLINE static btQuaternion LookAtYX(const btVector3& sourcePosition,const btVector3& targetPosition)	{
	btVector3 D(targetPosition-sourcePosition);
	btScalar Dxz2(D.x()*D.x()+D.z()*D.z());
	btScalar Dxz(btSqrt(Dxz2));
	btScalar AY(btAtan2(D.x(),D.z()));
	btScalar AX(-btAtan2(D.y(),Dxz));
	return btQuaternion(btVector3(0.,1.,0.),AY) * btQuaternion(btVector3(1.,0.,0.),AX);
}
void btRailwayCompoundShape::addPoints(const btVector3* ppoints,const int numPoints)	{
	if (numPoints<=0) return;
	int startIndex = 0;
	if (getNumChildShapes()==0) {
                firstT.setOrigin(ppoints[0]);
		T0.setOrigin(ppoints[0]);
                if (numPoints>1) {
                    T0.getBasis().setRotation(LookAtYX(ppoints[0],ppoints[1]));
                    firstT.setBasis(T0.getBasis());
                }
	}
	for (int i=startIndex;i<numPoints;i++) addPoint(ppoints[i]);
}
void btRailwayCompoundShape::addPoint(const btVector3& point,const btTransform* optionalT1)	{
	static const btScalar pointSquaredPrecision(0.01f);
	if ((point-T0.getOrigin()).length2()<pointSquaredPrecision) return;
	/*
			      VL1
	             /
	            /           T1
			   /           /          
			  /           /           VR1          
			 /           /           /
			/           /           /
			|			|			|
			VL0         T0          VR0
	*/
	btTransform T1;
        T1.getBasis().setRotation(LookAtYX(T0.getOrigin(),point));
        T1.setOrigin(point);

	const bool oddNumberOfSets(numOfRailwaySets%2==1);
        if (oddNumberOfSets) addPoint(T0,T1,optionalT1);
	
        const bool improvePrecision = true;
	for (int set=0,numHalfSets=(numOfRailwaySets-(oddNumberOfSets?1:0))/2;set<numHalfSets;set++)	{
		btTransform T00,T11;
		const btScalar distanceToSet = oddNumberOfSets ? (2*halfDistanceBetweenTwoRailwaySets+2*halfDistanceBetweenRails)*(set+1) : 
									(halfDistanceBetweenTwoRailwaySets+halfDistanceBetweenRails+(2*halfDistanceBetweenTwoRailwaySets+2*halfDistanceBetweenRails)*set);
                const btVector3 T1Column0(optionalT1?(optionalT1->getBasis().getColumn(0)):(T1.getBasis().getColumn(0)));
                // Left set
		{
			T00 = btTransform(T0.getBasis(),T0.getOrigin()+T0.getBasis().getColumn(0)*distanceToSet);
                        T11 = btTransform(T1.getBasis(),T1.getOrigin()+T1Column0*distanceToSet);
                        if (improvePrecision) T11.getBasis().setRotation(LookAtYX(T00.getOrigin(),T11.getOrigin()));  // improves precision
                        addPoint(T00,T11,optionalT1);
		}
		// Right set
		{
			T00 = btTransform(T0.getBasis(),T0.getOrigin()-T0.getBasis().getColumn(0)*distanceToSet);
                        T11 = btTransform(T1.getBasis(),T1.getOrigin()-T1Column0*distanceToSet);
                        if (improvePrecision) T11.getBasis().setRotation(LookAtYX(T00.getOrigin(),T11.getOrigin()));  // improves precision
                        addPoint(T00,T11,optionalT1);
		}		
		
		
	}	

	
	T0 = T1;
}
void btRailwayCompoundShape::addPoint(const btTransform& T0,const btTransform& T1,const btTransform* optionalT1)	{
	const btVector3 VL0(T0.getOrigin()+T0.getBasis().getColumn(0)*halfDistanceBetweenRails);
	const btVector3 VR0(T0.getOrigin()-T0.getBasis().getColumn(0)*halfDistanceBetweenRails);

        const btVector3 T1Column0(optionalT1?(optionalT1->getBasis().getColumn(0)):(T1.getBasis().getColumn(0)));
        const btVector3 VL1(T1.getOrigin()+T1Column0*halfDistanceBetweenRails);
        const btVector3 VR1(T1.getOrigin()-T1Column0*halfDistanceBetweenRails);
		
	btTransform TL(T1.getBasis(),(VL0+VL1)*0.5f);
	btTransform TR(T1.getBasis(),(VR0+VR1)*0.5f);
	
	TL.getOrigin()=btVector3(TL.getOrigin()+TL.getBasis().getColumn(1)*railwayHalfHeight);
	TR.getOrigin()=btVector3(TR.getOrigin()+TR.getBasis().getColumn(1)*railwayHalfHeight);
	
	const btVector3 boxLeftHalfExtents(singleRailHalfWidth,railwayHalfHeight,(VL1-VL0).length()*0.5f);
	const btVector3 boxRightHalfExtents(singleRailHalfWidth,railwayHalfHeight,(VR1-VR0).length()*0.5f);
	
	btBoxShape* boxLeftShape(NULL);
	btBoxShape* boxRightShape(NULL);

	// Box shape caching and creating code
	{
		const int numRailwayShapes(railwayShapes.size());
                for (int i = numRailwayShapes-1;(i>=0 && i>numRailwayShapes-1-cacheBoxShapesFactor);i--)	{
			const btBoxShape* shape(railwayShapes[i]);
			if (shape)	{
				const btVector3 halfExtents = shape->getHalfExtentsWithMargin();//shape->getHalfExtentsWithMargin(); or shape->getHalfExtentsWithoutMargin() ?
				if (!boxLeftShape && (halfExtents-boxLeftHalfExtents).length2()<cacheBoxShapesSquaredPrecision) boxLeftShape = const_cast < btBoxShape* > (shape);
				if (!boxRightShape && (halfExtents-boxRightHalfExtents).length2()<cacheBoxShapesSquaredPrecision) boxRightShape = const_cast < btBoxShape* > (shape);
				if (boxLeftShape && boxRightShape) break;
			}
		}
		if (!boxLeftShape) {
			boxLeftShape = new btBoxShape(boxLeftHalfExtents);
			railwayShapes.push_back(boxLeftShape);
			#ifdef USE_BTTRIANGLEMESHEX
			if (myApp && referenceBoxShapedMesh) {
				myApp->AssignMeshCopy(boxLeftShape,referenceBoxShapedMesh);
				myApp->GetMeshOverrides(*boxLeftShape).scaling = boxLeftShape->getHalfExtentsWithMargin()/referenceBoxShapedMeshHalfExtents;
			}
			#endif //USE_BTTRIANGLEMESHEX
			++numOfCreatedDifferentBoxShapes;
		}	
		if (!boxRightShape)	{
			if ((boxLeftHalfExtents-boxRightHalfExtents).length2()<cacheBoxShapesSquaredPrecision) boxRightShape = boxLeftShape;
			else {
				boxRightShape = new btBoxShape(boxRightHalfExtents);
				railwayShapes.push_back(boxRightShape);
				#ifdef USE_BTTRIANGLEMESHEX
				if (myApp && referenceBoxShapedMesh) {
					myApp->AssignMeshCopy(boxRightShape,referenceBoxShapedMesh);
					myApp->GetMeshOverrides(*boxRightShape).scaling = boxRightShape->getHalfExtentsWithMargin()/referenceBoxShapedMeshHalfExtents;
				}	
				#endif //USE_BTTRIANGLEMESHEX									
				++numOfCreatedDifferentBoxShapes;
			}	
		}
	}		

		
	addChildShape(TL,boxLeftShape);
	addChildShape(TR,boxRightShape);
	
}
void btRailwayCompoundShape::connectEndToStart()	{
	if (getNumChildShapes()<=2) return;
        addPoint(firstT.getOrigin(),&firstT);
}
void btRailwayCompoundShape::displayDebugInfo() const	{
	printf("btRailwayCompoundShape::numOfCreatedDifferentBoxShapes=%d/%d\n",numOfCreatedDifferentBoxShapes,getNumChildShapes());
}

void btRailwayCompoundShape::AssignBoxMeshForRailway(class GlutDemoApplicationEx* _myApp,const btTriangleMeshEx* mesh,const bool useFlatNormals,const bool avoidGlNormalizeCallsWhenScalingMesh)	{
#ifdef USE_BTTRIANGLEMESHEX
		if (_myApp && mesh) {
			myApp = _myApp;
			referenceBoxShapedMeshHalfExtents = mesh->getAabbHalfExtents();
			const btScalar fuzzyZero(0.000001f);
			if (referenceBoxShapedMeshHalfExtents.x()<fuzzyZero) referenceBoxShapedMeshHalfExtents.setX(fuzzyZero);
			if (referenceBoxShapedMeshHalfExtents.y()<fuzzyZero) referenceBoxShapedMeshHalfExtents.setY(fuzzyZero);
			if (referenceBoxShapedMeshHalfExtents.z()<fuzzyZero) referenceBoxShapedMeshHalfExtents.setZ(fuzzyZero);			
			if (referenceBoxShapedMesh)	{delete referenceBoxShapedMesh;referenceBoxShapedMesh=NULL;}
			referenceBoxShapedMesh = new btBoxShape(referenceBoxShapedMeshHalfExtents);
			myApp->AssignMesh(referenceBoxShapedMesh,*mesh,useFlatNormals);
                        myApp->GetMeshOverrides(*referenceBoxShapedMesh).avoidGlNormalizeCallsWhenScaling=avoidGlNormalizeCallsWhenScalingMesh;
		
			for (int i=0,sz=railwayShapes.size();i<sz;i++)	{
				btBoxShape* shape = railwayShapes[i];
				if (!shape) continue;
				myApp->AssignMeshCopy(shape,referenceBoxShapedMesh);
				myApp->GetMeshOverrides(*shape).scaling = shape->getHalfExtentsWithMargin()/referenceBoxShapedMeshHalfExtents;
			}
		}
#endif //USE_BTTRIANGLEMESHEX		
}
void btRailwayCompoundShape::AssignDefaultBoxMeshForRailway(class GlutDemoApplicationEx* _myApp,const bool useFlatNormals, const int optionalTextureId,const float optionalTextureRepX,const float optionalTextureRepY,const float optionalTextureOffsetX,const float optionalTextureOffsetY)    {
#ifdef USE_BTTRIANGLEMESHEX
    if (_myApp) {
        btTriangleMeshEx mesh;
        if (mesh.loadDefaultBoxMesh(btVector3(singleRailHalfWidth,railwayHalfHeight,15),(optionalTextureId>0),false,optionalTextureRepX,optionalTextureRepY,optionalTextureOffsetX,optionalTextureOffsetY,3))	{
                // Line above: last arg: 3 faces only (and are the correct ones (+Y,+X and -X)

                Material mat(0.424802f,0.366001f,0.421578f,true,0.5f,0.35f);
                mesh.setMeshSubPartMaterial(0,mat);	// Note that when setting a material directly to a mesh it doesn't need to be persistent (when setting it as a mesh override it does).
                if (optionalTextureId>0) mesh.setMeshSubPartTextureId(0,optionalTextureId);

                AssignBoxMeshForRailway(_myApp,&mesh,useFlatNormals,true);
        }
    }
#endif //USE_BTTRIANGLEMESHEX
}

#undef USE_BTTRIANGLEMESHEX

