/*
This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/
#pragma once
#ifndef BTWHEEL_SHAPES_H
#define BTWHEEL_SHAPES_H

#include "BulletCollision/CollisionShapes/btMultiSphereShape.h"
#include "BulletCollision/CollisionShapes/btCompoundShape.h"
#include "BulletCollision/CollisionShapes/btCylinderShape.h"


/** A collision shape to model a convex tyre shape.
    The axis of rotation (wheel height) is in the X direction.
	This is just a wrapper around btCylinderShapeX, so that I can pass just 2 args to its ctr
	(my laziness rules...).

    -------------------
    LAST MINUTE UPDATE: As RBD showed in one of his YouTube videos, we can dramatically improve this shape by
    -------------------
    enlarging its collision margin, so that it looks like a real tyre. This is both high quality and efficient!
    To test this approach, simply change the default value of _setExtendedCollisionMargin in the ctr to 'true'.

    I leave it to 'false' ATM, because it's not perfectly clear to me what's the best value to use based on the shape AABB:
    please see the implementation of setExtendedCollisionMargin() in the cpp file.
    Also, if we need to use setLocalScaling(), it should probably be overridden to scale the collision margin too.
*/
class btWheelCylinderShape : public btCylinderShapeX {
	protected:
	
	public:
    btWheelCylinderShape(const btScalar radius,const btScalar height,bool _setExtendedCollisionMargin=false)
	: btCylinderShapeX(btVector3(height*0.5f,radius,radius))
    {
        if (_setExtendedCollisionMargin) setExtendedCollisionMargin();
    }
    void setExtendedCollisionMargin();
	virtual ~btWheelCylinderShape() {}
};

/** A collision shape to model a convex tyre shape.
    The axis of rotation (wheel height) is in the X direction.
   	This is just a single scaled sphere!
	The advantage is that there are no "segmented circles" inside it, and that it's very light.
*/
class btWheelSimpleShape : public btMultiSphereShape {
	protected:
    typedef btMultiSphereShape base;

	public:
	btWheelSimpleShape(const btScalar radius,const btScalar height) 
	: btMultiSphereShape(*Create(radius,height)),newLocalScaling(1,1,1)
	{}
	virtual ~btWheelSimpleShape() {}

	virtual const char*	getName()const {return "btWheelSimpleShape";}

    virtual void	calculateLocalInertia(btScalar mass,btVector3& inertia) const;

	// Local Scaling Override------------
	typedef btMultiSphereShape Base;
	virtual void	setLocalScaling(const btVector3& scaling)	{
		btVector3 baseLocalScaling = Base::getLocalScaling();	// It's baseLocalScaling * newLocalScaling
		if (newLocalScaling.x()==0) newLocalScaling.setX(SIMD_EPSILON);
		if (newLocalScaling.y()==0) newLocalScaling.setY(SIMD_EPSILON);
		if (newLocalScaling.z()==0) newLocalScaling.setZ(SIMD_EPSILON);	
		baseLocalScaling = baseLocalScaling / newLocalScaling;
				
		newLocalScaling = scaling;
		Base::setLocalScaling(newLocalScaling * baseLocalScaling);	
	}
	virtual const btVector3& getLocalScaling() const	{
		return newLocalScaling;
	}
	protected:
	btVector3 newLocalScaling;
	//------------------------------------


	private:
	
	static btMultiSphereShape* Create(const btScalar radius,const btScalar height);	
};

/** A collision shape to model a convex tyre shape.
    The axis of rotation (wheel height) is in the X direction.
	This is a more complex btMultiSphereShape (that looks like a real wheel)
	The main "circular section" in the center isn't segmented, but the two slightly smaller "circular sections" on the sides are "segmented".
	if _smallSpheresHeightFactor == 1.0f, the small "circular sections" at the sides have the same radius as the bigger one at the center.	    
	Warning: The shape might be a bit bigger than needed (To be fixed?).
*/
class btWheelShape : public btMultiSphereShape {
	protected:
    typedef btMultiSphereShape base;

	public:
	btWheelShape(const btScalar radius,const btScalar height,const int numSegments=16,const btScalar _smallSpheresHeightFactor=0.975f) 
	: btMultiSphereShape(*Create(radius,height,numSegments,_smallSpheresHeightFactor)),newLocalScaling(1,1,1)
	{}
	virtual ~btWheelShape() {}
	virtual const char*	getName()const {return "btWheelShape";}

    virtual void	calculateLocalInertia(btScalar mass,btVector3& inertia) const;

    // Local Scaling Override------------
	typedef btMultiSphereShape Base;
	virtual void	setLocalScaling(const btVector3& scaling)	{
		btVector3 baseLocalScaling = Base::getLocalScaling();	// It's baseLocalScaling * newLocalScaling
		if (newLocalScaling.x()==0) newLocalScaling.setX(SIMD_EPSILON);
		if (newLocalScaling.y()==0) newLocalScaling.setY(SIMD_EPSILON);
		if (newLocalScaling.z()==0) newLocalScaling.setZ(SIMD_EPSILON);	
		baseLocalScaling = baseLocalScaling / newLocalScaling;
				
		newLocalScaling = scaling;
		Base::setLocalScaling(newLocalScaling * baseLocalScaling);	
	}
	virtual const btVector3& getLocalScaling() const	{
		return newLocalScaling;
	}
	protected:
	btVector3 newLocalScaling;
	//------------------------------------
	
	private:
	
	static btMultiSphereShape* Create(const btScalar radius,const btScalar height,const int numSegments=16,const btScalar _smallSpheresHeightFactor=0.95f);
};


//#define BTWHEENTRAINSHAPE_UPCAST_USES_DYNAMIC_CAST
#ifndef BTWHEENTRAINSHAPE_UPCAST_USES_DYNAMIC_CAST
#include <string.h> // just for strmp
#endif // BTWHEENTRAINSHAPE_UPCAST_USES_DYNAMIC_CAST

/** A collision shape to model a simple train wheel that can be used to "ride" narrow box shapes.
    The axis of rotation (named wheel width or wheel height) is in the X direction.
	This is a btCompoundShape made with a btCylinderX at the center plus 2 cones at the sides with their base outside.
	|\ - /|
	| | | |
	|/ - \|
	Because of this the total width must be longer than the height of one cone (so that its vertex does not come out of the other side).
	That's why the ctr takes an argument called: totWidthThatCanBeIncreased.
	P.S. It's surely possible to delete this requirement, but this code is old, and I'm very lazy... do it better yourself!
*/
class btWheelTrainShape : public btCompoundShape {
	protected:
    typedef btCompoundShape base;

	public:
	/** totRadius = total radius of the wheel
	    totWidthThatCanBeIncreased = total width of the wheel
	*/
	btWheelTrainShape(const btScalar totRadius,const btScalar totWidthThatCanBeIncreased,const btScalar innerRadius=btScalar(-1.f),const btScalar innerWidth=btScalar(-1.f)) 
	: btCompoundShape(*Create(totRadius,totWidthThatCanBeIncreased,innerRadius,innerWidth))
	{}
	virtual ~btWheelTrainShape() {}
	virtual const char*	getName()const {return "btWheelTrainShape";}

    virtual void	calculateLocalInertia(btScalar mass,btVector3& inertia) const;

    SIMD_FORCE_INLINE static btWheelTrainShape* upcast(btCollisionShape* shape) {
        #ifdef BTWHEENTRAINSHAPE_UPCAST_USES_DYNAMIC_CAST
        return dynamic_cast <btWheelTrainShape*> (shape);
        #else
        if (!shape || shape->getShapeType()!=COMPOUND_SHAPE_PROXYTYPE) return NULL;
        btCompoundShape* cs = static_cast < btCompoundShape* > (shape);
        if (cs && cs->getNumChildShapes()==3 && strcmp(shape->getName(),"btWheelTrainShape")==0) return static_cast<btWheelTrainShape*>(shape);
        return NULL;
        #endif
    }
    SIMD_FORCE_INLINE static const btWheelTrainShape* upcast(const btCollisionShape* shape) {
         #ifdef BTWHEENTRAINSHAPE_UPCAST_USES_DYNAMIC_CAST
        return dynamic_cast <const btWheelTrainShape*> (shape);
        #else
        if (!shape || shape->getShapeType()!=COMPOUND_SHAPE_PROXYTYPE) return NULL;
        const btCompoundShape* cs = static_cast < const btCompoundShape* > (shape);
        if (cs && cs->getNumChildShapes()==3 && strcmp(shape->getName(),"btWheelTrainShape")==0) return static_cast<const btWheelTrainShape*>(shape);
        return NULL;
        #endif
    }

	private:
	
	static btCompoundShape* Create(btScalar outerRadius,btScalar totalWidth,btScalar _innerRadius,btScalar _innerWidth);
};


/** A class that should be used to build a simple planar railway.
    It should be better to unwrap it for static usage (AFAIK).
*/
class btRailwayCompoundShape : public btCompoundShape	{
	public:
	/**
	    ALL THE MEASURES ARE TAKEN IN THIS WAY (top view of 2 railway sets):
	    
	    singleRailFullWidth
		  |
		|--|
		|  |         |  |                       |  |         |  |      
		|  |         |  |                       |  |         |  |
		|  |         |  |                       |  |         |  |  
		|  |         |  |                       |  |         |  | 
		|  |         |  |                       |  |         |  |
		|--|          |--------------------------|
          |-----------|             |
                |                   | 
        fullDistanceBetweenRails    |
		               distanceBetweenTwoRailwaySets
	*/
        btRailwayCompoundShape(const btScalar fullDistanceBetweenRails,const btScalar singleRailFullWidth,const btScalar railwayHeight,const unsigned short _numOfParalleleRailwaySets=1,const btScalar distanceBetweenTwoRailwaySets=-1,const int _cacheBoxShapesFactor=50,const btScalar _cacheBoxShapesSquaredPrecision=btScalar(0.01))
            : btCompoundShape(),halfDistanceBetweenRails(fullDistanceBetweenRails*0.5),singleRailHalfWidth(singleRailFullWidth*0.5),railwayHalfHeight(railwayHeight*0.5),
	numOfRailwaySets(_numOfParalleleRailwaySets),halfDistanceBetweenTwoRailwaySets(distanceBetweenTwoRailwaySets*0.5),numOfCreatedDifferentBoxShapes(0),
        cacheBoxShapesFactor(_cacheBoxShapesFactor>=0?_cacheBoxShapesFactor:0),cacheBoxShapesSquaredPrecision(_cacheBoxShapesSquaredPrecision),referenceBoxShapedMesh(NULL),myApp(NULL),referenceBoxShapedMeshHalfExtents(1,1,1)
	{
		if (halfDistanceBetweenTwoRailwaySets<=0) halfDistanceBetweenTwoRailwaySets = halfDistanceBetweenRails*0.75;
		if (numOfRailwaySets==0) numOfRailwaySets=1;
		T0.setIdentity();
                firstT.setIdentity();
	}
	virtual ~btRailwayCompoundShape() {
		if (referenceBoxShapedMesh)	{
			delete referenceBoxShapedMesh;referenceBoxShapedMesh=NULL;
		}
	}
	virtual const char*	getName()const {return "btRailwayCompoundShape";}
	
	void addPoints(const btVector3* ppoints,const int numPoints);	// Can be called multiple times, but must be: numPoints>=2.
	void connectEndToStart();	// It should be better to call this at the end
	
	void displayDebugInfo() const;
	
        void AssignBoxMeshForRailway(class GlutDemoApplicationEx* _myApp,const class btTriangleMeshEx* mesh,const bool useFlatNormals=false,const bool avoidGlNormalizeCallsWhenScalingMesh=false);
        void AssignDefaultBoxMeshForRailway(class GlutDemoApplicationEx* _myApp,const bool useFlatNormals=false,const int optionalTextureId=0,const float optionalTextureRepX=1,const float optionalTextureRepY=1,const float optionalTextureOffsetX=0,const float optionalTextureOffsetY=0);

	protected:
	btScalar halfDistanceBetweenRails;
	btScalar singleRailHalfWidth;
	btScalar railwayHalfHeight;
	btTransform T0;	// current transform (global space)
        btTransform firstT;	// needed to close the loop
        void addPoint(const btVector3& P,const btTransform* optionalT1=NULL);   // optionalT1 is just needed to better close the loop
        void addPoint(const btTransform& T0,const btTransform& T1,const btTransform* optionalT1=NULL);   // optionalT1 is just needed to better close the loop
	//Extended:
	btScalar halfDistanceBetweenTwoRailwaySets;
	unsigned short numOfRailwaySets;
	
        int cacheBoxShapesFactor;
	btScalar cacheBoxShapesSquaredPrecision;
	unsigned int numOfCreatedDifferentBoxShapes;	//Debug
	
	btAlignedObjectArray<btBoxShape*> railwayShapes;
	
	btBoxShape* referenceBoxShapedMesh;	// Just to contain mesh display lists, must be deleted by dctr
	btVector3 referenceBoxShapedMeshHalfExtents;
	class GlutDemoApplicationEx* myApp;
};



#endif //BTWHEEL_SHAPES_H
